// $Id: PnfsHandler.java,v 1.35 2007-10-14 01:51:47 behrmann Exp $

package diskCacheV111.util ;

import java.util.List;
import java.util.Collections;
import java.util.Set;
import java.util.EnumSet;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import dmg.cells.nucleus.CellEndpoint;
import dmg.cells.nucleus.CellPath;
import dmg.cells.nucleus.NoRouteToCellException;

import diskCacheV111.vehicles.PnfsAddCacheLocationMessage;
import diskCacheV111.vehicles.PnfsClearCacheLocationMessage;
import diskCacheV111.vehicles.PnfsCreateDirectoryMessage;
import diskCacheV111.vehicles.PnfsCreateEntryMessage;
import diskCacheV111.vehicles.PnfsDeleteEntryMessage;
import diskCacheV111.vehicles.PnfsFlagMessage;
import diskCacheV111.vehicles.PnfsGetCacheLocationsMessage;
import diskCacheV111.vehicles.PnfsGetStorageInfoMessage;
import diskCacheV111.vehicles.PnfsMapPathMessage;
import diskCacheV111.vehicles.PnfsMessage;
import diskCacheV111.vehicles.PnfsRenameMessage;
import diskCacheV111.vehicles.PnfsSetFileMetaDataMessage;
import diskCacheV111.vehicles.PnfsSetLengthMessage;
import diskCacheV111.vehicles.PnfsSetStorageInfoMessage;
import diskCacheV111.vehicles.PnfsGetParentMessage;
import diskCacheV111.vehicles.PnfsSetChecksumMessage;
import diskCacheV111.vehicles.PoolFileFlushedMessage;
import diskCacheV111.vehicles.StorageInfo;

import org.dcache.cells.CellMessageSender;
import org.dcache.cells.CellStub;
import org.dcache.auth.Subjects;
import org.dcache.namespace.FileAttribute;
import org.dcache.vehicles.FileAttributes;
import org.dcache.vehicles.PnfsGetFileAttributes;
import org.dcache.vehicles.PnfsSetFileAttributes;
import org.dcache.namespace.FileType;
import org.dcache.acl.enums.AccessMask;
import org.dcache.util.Checksum;

import javax.security.auth.Subject;

import static org.dcache.namespace.FileAttribute.*;

public class PnfsHandler
    implements CellMessageSender
{
    private final String _poolName;
    private static final long DEFAULT_PNFS_TIMEOUT = 30 * 60 * 1000L;

    private final CellStub _cellStub;

    private Subject _subject;

    private static final Logger _logNameSpace =
        LoggerFactory.getLogger("logger.org.dcache.namespace." + PnfsHandler.class.getName());

    private static CellStub createStub(CellPath path)
    {
        CellStub stub = new CellStub();
        stub.setDestinationPath(path);
        stub.setTimeout(DEFAULT_PNFS_TIMEOUT);
        stub.setRetryOnNoRouteToCell(true);
        return stub;
    }

    @Deprecated
    public PnfsHandler(CellEndpoint endpoint,
                       CellPath pnfsManagerPath,
                       String poolName)
    {
        this(pnfsManagerPath, poolName);
        setCellEndpoint(endpoint);
    }

    @Deprecated
    public PnfsHandler(CellEndpoint endpoint,
                       CellPath pnfsManagerPath)
    {
        this(pnfsManagerPath);
        setCellEndpoint(endpoint);
    }

    public PnfsHandler(CellPath pnfsManagerPath)
    {
        this(createStub(pnfsManagerPath));
    }

    public PnfsHandler(CellPath pnfsManagerPath,
                       String poolName)
    {
        this(createStub(pnfsManagerPath), poolName);
    }

    public PnfsHandler(CellStub stub)
    {
        this(stub, "<client>");
    }

    public PnfsHandler(CellStub stub, String poolName)
    {
        _cellStub = stub;
        _poolName = poolName;
    }

    /**
     * Copy constructor. The primary purpose is to create session
     * specific PnfsHandlers with a session specific subject. Notice
     * that the CellStub is shared between the two handlers and thus
     * the timeout will always be the same.
     *
     * @param handler The PnfsHandler to copy
     * @param subject The Subject to apply to the copy
     */
    public PnfsHandler(PnfsHandler handler, Subject subject)
    {
        _poolName = handler._poolName;
        _cellStub = handler._cellStub;
        _subject = subject;
    }

    @Override
    public void setCellEndpoint(CellEndpoint endpoint)
    {
        _cellStub.setCellEndpoint(endpoint);
    }

    public void setSubject(Subject subject)
    {
        _subject = subject;
    }

    /**
     * Sends a PnfsMessage to PnfsManager.
     *
     * @throws NoRouteToCellException if the PnfsManager could not be reached
     */
    public void send(PnfsMessage msg)
        throws NoRouteToCellException
    {
        if (_cellStub == null) {
            throw new IllegalStateException("Missing endpoint");
        }

        if (_subject != null) {
            msg.setSubject(_subject);
        }

        _cellStub.send(msg);
    }

    /**
     * Sends a PnfsMessage notification to PnfsManager. No reply is
     * expected for a notification and no failure is reported if the
     * message could not be delivered.
     */
    public void notify(PnfsMessage msg)
    {
        try {
            msg.setReplyRequired(false);
            send(msg);
        } catch (NoRouteToCellException e) {
            _logNameSpace.warn("Failed to deliver message " +
                               msg.getClass().getSimpleName() +
                               " to PnfsManager: " + e.getMessage());
        }
    }

   public void clearCacheLocation( PnfsId pnfsId ){
       clearCacheLocation( pnfsId , false  );
   }
   public void clearCacheLocation( PnfsId pnfsId , boolean removeIfLast  ){

       notify( new PnfsClearCacheLocationMessage(
                           pnfsId,
                           _poolName,
                           removeIfLast)
           ) ;

   }
   public void clearCacheLocation( PnfsId pnfsId , String poolName ){

       notify( new PnfsClearCacheLocationMessage(
                           pnfsId,
                           poolName)
           ) ;

   }

   public void addCacheLocation( PnfsId pnfsId )
       throws CacheException
   {
       pnfsRequest( new PnfsAddCacheLocationMessage(
                           pnfsId,
                           _poolName)
           ) ;

   }

   public void setFileSize( PnfsId pnfsId , long length )throws CacheException {

       pnfsRequest( new PnfsSetLengthMessage(
                           pnfsId,
                           length  )
           ) ;

   }
   public List<String> getCacheLocations( PnfsId pnfsId )throws CacheException {
      PnfsGetCacheLocationsMessage pnfsMessage = new PnfsGetCacheLocationsMessage(pnfsId) ;
      pnfsMessage = pnfsRequest(pnfsMessage) ;
      List<String> assumedLocations = pnfsMessage.getCacheLocations() ;

      if (assumedLocations == null) {
          return Collections.emptyList();
      } else {
          return assumedLocations;
      }
   }

   public List<String> getCacheLocationsByPath( String fileName )throws CacheException {
      PnfsGetCacheLocationsMessage pnfsMessage = new PnfsGetCacheLocationsMessage() ;
      pnfsMessage.setPnfsPath( fileName ) ;
      pnfsMessage = pnfsRequest(pnfsMessage) ;
      List<String> assumedLocations = pnfsMessage.getCacheLocations() ;

      if (assumedLocations == null) {
          return Collections.emptyList();
      } else {
          return assumedLocations;
      }
   }

    /**
     * Sends a message to the request manager and blocks until a reply
     * is received. In case of errors in the reply, those are thrown
     * as a CacheException. Timeouts and failure to send the message
     * to the PnfsManager are reported as a timeout CacheException.
     */
   public <T extends PnfsMessage> T pnfsRequest( T msg )
           throws CacheException {

       if (_cellStub == null)
           throw new IllegalStateException("Missing endpoint");

        try {
            msg.setReplyRequired(true);
            if (_subject != null) {
                msg.setSubject(_subject);
            }
            return _cellStub.sendAndWait(msg);
        } catch (InterruptedException e) {
            throw  new CacheException(CacheException.UNEXPECTED_SYSTEM_EXCEPTION,
                    "Sending message to PnafsManager intterupted");
        }
   }

    public PnfsCreateEntryMessage createPnfsDirectory(String path)
        throws CacheException
    {
        return pnfsRequest(new PnfsCreateDirectoryMessage(path));
    }

    public PnfsCreateEntryMessage createPnfsDirectory(String path, int uid, int gid, int mode)
        throws CacheException
    {
        return pnfsRequest(new PnfsCreateDirectoryMessage(path, uid, gid, mode));
    }

    /**
     * Creates a directory and all its parent directories.
     *
     * REVISIT: Should eventually be moved to PnfsManager with a flag
     * in the PnfsCreateEntryMessage indicating whether parent
     * directories should be created.
     *
     * @returns the FileAttributes of <code>path</code>
     */
    public FileAttributes createDirectories(FsPath path)
        throws CacheException
    {
        PnfsCreateEntryMessage message;
        try {
            message = createPnfsDirectory(path.toString());
        } catch (FileNotFoundCacheException e) {
            createDirectories(path.getParent());
            message = createPnfsDirectory(path.toString());
        }

        /* In case of incomplete create, delete the directory right
         * away.
         */
        if (message.getStorageInfo() == null) {
            try {
                deletePnfsEntry(message.getPnfsId(), path.toString());
            } catch (FileNotFoundCacheException e) {
                // Already gone, so never mind
            } catch (CacheException e) {
                _logNameSpace.error(e.toString());
            }

            throw new CacheException("Failed to create directory: " + path);
        }

        return message.getFileAttributes();
    }

   public PnfsCreateEntryMessage createPnfsEntry( String path )
          throws CacheException                {

       return pnfsRequest(new PnfsCreateEntryMessage( path )) ;

   }

    public void pnfsSetFileMetaData(PnfsId pnfsId, FileMetaData meta)
        throws CacheException
    {
        pnfsRequest(new PnfsSetFileMetaDataMessage(pnfsId, meta));
    }

    public void renameEntry(PnfsId pnfsId, String newName)
        throws CacheException
    {
        renameEntry(pnfsId, newName, true);
    }

    public void renameEntry(PnfsId pnfsId, String newName, boolean overwrite)
        throws CacheException
    {
        pnfsRequest(new PnfsRenameMessage(pnfsId, newName, overwrite));
    }

    public void renameEntry(String path, String newName, boolean overwrite)
        throws CacheException
    {
        pnfsRequest(new PnfsRenameMessage(path, newName, overwrite));
    }

   public PnfsCreateEntryMessage createPnfsEntry( String path , int uid , int gid , int mode )
          throws CacheException                {

       return pnfsRequest( new PnfsCreateEntryMessage( path , uid , gid , mode ) ) ;

   }

   public void setStorageInfoByPnfsId(
                PnfsId pnfsId , StorageInfo  storageInfo , int accessMode )
          throws CacheException                {

       /*
        * use request, which throws exceptions in case of.....
        */
       pnfsRequest( new PnfsSetStorageInfoMessage( pnfsId , storageInfo , accessMode ) ) ;
   }
   public PnfsGetStorageInfoMessage getStorageInfoByPnfsId( PnfsId pnfsId )
          throws CacheException                {
      return  pnfsRequest(new PnfsGetStorageInfoMessage( pnfsId )) ;

   }

   public PnfsGetStorageInfoMessage getStorageInfoByPath( String pnfsPath )
          throws CacheException                {
      return getStorageInfoByPath(pnfsPath, false) ;

   }

   public PnfsGetStorageInfoMessage getStorageInfoByPath( String pnfsPath ,
       boolean requestChecksum)
          throws CacheException                {

      PnfsGetStorageInfoMessage sInfo = new PnfsGetStorageInfoMessage() ;
      sInfo.setPnfsPath( pnfsPath ) ;
      sInfo.setChecksumsRequested(requestChecksum);
      return pnfsRequest( sInfo ) ;

   }

    public PnfsId getParentOf(PnfsId pnfsId)
        throws CacheException
    {
            return pnfsRequest(new PnfsGetParentMessage(pnfsId)).getParent();
    }

    public void deletePnfsEntry(String path) throws CacheException
    {
        deletePnfsEntry(null, path);
    }

    public void deletePnfsEntry(String path, Set<FileType> allowed)
        throws CacheException
    {
        deletePnfsEntry(null, path, allowed);
    }

    public void deletePnfsEntry(PnfsId pnfsid)  throws CacheException
    {
        deletePnfsEntry(pnfsid, null);
    }

    public void deletePnfsEntry(PnfsId pnfsid, String path)
        throws CacheException
    {
        deletePnfsEntry(pnfsid, path, EnumSet.allOf(FileType.class));
    }

    public void deletePnfsEntry(PnfsId pnfsid, String path,
                                Set<FileType> allowed)
        throws CacheException
    {
        pnfsRequest(new PnfsDeleteEntryMessage(pnfsid, path, allowed));
    }

   /**
    * Getter for property __pnfsTimeout.
    * @return Value of property __pnfsTimeout.
    */
   public long getPnfsTimeout() {
       return _cellStub.getTimeout();
   }

   /**
    * Setter for property __pnfsTimeout.
    * @param __pnfsTimeout New value of property __pnfsTimeout.
    */
   public void setPnfsTimeout(long pnfsTimeout) {
       _cellStub.setTimeout(pnfsTimeout);
   }

   public String getPnfsFlag(PnfsId pnfsId, String flag)
   throws CacheException
   {
       PnfsFlagMessage flagMessage =
                new PnfsFlagMessage( pnfsId ,flag , PnfsFlagMessage.FlagOperation.GET ) ;
       flagMessage.setReplyRequired( true );

       return pnfsRequest(flagMessage).getValue();
   }

   public void putPnfsFlag(PnfsId pnfsId, String flag, String value)
   {
       PnfsFlagMessage flagMessage =
                new PnfsFlagMessage( pnfsId ,flag , PnfsFlagMessage.FlagOperation.SET ) ;
       flagMessage.setReplyRequired( false );
       flagMessage.setValue(value);
       notify(flagMessage);
   }

   public void fileFlushed(PnfsId pnfsId, StorageInfo storageInfo ) throws CacheException {

	   PoolFileFlushedMessage fileFlushedMessage = new PoolFileFlushedMessage(_poolName, pnfsId, storageInfo);

	   // throws exception if something goes wrong
	   pnfsRequest(fileFlushedMessage);

   }

	/**
	 * Get path corresponding to given pnfsid.
	 *
	 * @param pnfsID
	 * @return path
	 * @throws CacheException
	 */
	public String getPathByPnfsId(PnfsId pnfsID) throws CacheException {
		return pnfsRequest(new PnfsMapPathMessage(pnfsID)).getGlobalPath();
	}

	/**
	 * Get pnfsid corresponding to given path.
	 *
	 * @param path
	 * @return pnfsid
	 * @throws CacheException
	 */
	public PnfsId getPnfsIdByPath(String path) throws CacheException {
		return pnfsRequest(new PnfsMapPathMessage(path)).getPnfsId();
	}

    /**
     * Get file attributes. The PnfsManager is free to return less attributes
     * than requested. If <code>attr</code> is an empty array, file existence
     * if checked.
     *
     * @param pnfsid
     * @param attr array of requested attributes.
     * @return requested attributes
     */
    public FileAttributes getFileAttributes(PnfsId pnfsid, Set<FileAttribute> attr) throws CacheException {
        return pnfsRequest(new PnfsGetFileAttributes(pnfsid, attr)).getFileAttributes();
    }

    /**
     * Get file attributes. The PnfsManager is free to return less attributes
     * than requested. If <code>attr</code> is an empty array, file existence
     * if checked.
     *
     * @param pnfsid
     * @param attr array of requested attributes.
     * @param mask Additional AccessMask access rights to check
     * @return requested attributes
     */
    public FileAttributes getFileAttributes(PnfsId pnfsid, Set<FileAttribute> attr, Set<AccessMask> mask)
        throws CacheException
    {
        PnfsGetFileAttributes msg = new PnfsGetFileAttributes(pnfsid, attr);
        msg.setAccessMask(mask);
        return pnfsRequest(msg).getFileAttributes();
    }

    /**
     * Get file attributes. The PnfsManager is free to return less attributes
     * than requested. If <code>attr</code> is an empty array, file existence
     * if checked.
     *
     * @param path
     * @param attr array of requested attributes.
     * @return requested attributes
     */
    public FileAttributes getFileAttributes(String path, Set<FileAttribute> attr)
        throws CacheException
    {
        return pnfsRequest(new PnfsGetFileAttributes(path, attr)).getFileAttributes();
    }

    /**
     * Get file attributes. The PnfsManager is free to return less attributes
     * than requested. If <code>attr</code> is an empty array, file existence
     * if checked.
     *
     * @param path
     * @param attr array of requested attributes.
     * @param mask Additional AccessMask access rights to check
     * @return requested attributes
     */
    public FileAttributes getFileAttributes(String path,
                                            Set<FileAttribute> attr,
                                            Set<AccessMask> mask)
        throws CacheException
    {
        PnfsGetFileAttributes msg = new PnfsGetFileAttributes(path, attr);
        msg.setAccessMask(mask);
        return pnfsRequest(msg).getFileAttributes();
    }

    /**
     * Set file attributes. If <code>attr</code> is an empty array,
     * file existence if checked.
     *
     * @param pnfsid
     * @param attr array of requested attributes.
     * @return requested attributes
     */
    public void setFileAttributes(PnfsId pnfsid, FileAttributes attr)
        throws CacheException
    {
        pnfsRequest(new PnfsSetFileAttributes(pnfsid, attr));
    }

    public void setChecksum(PnfsId pnfsId, Checksum checksum)
        throws CacheException
    {
        PnfsSetChecksumMessage message =
            new PnfsSetChecksumMessage(pnfsId,
                                       checksum.getType().getType(),
                                       checksum.getValue());
        pnfsRequest(message);
    }
}
